#include "MCEM.h"
#define REPEAT 1

MCEM::MCEM(PriorParam p,Args a,vector<string> l,int i):
				iteration(i),
				prior(p),opflist(l),args(a){
	pathraw="";
	if (args.nargs()==4) pathraw=args.arg(3);
	for(int i=0;i<5;i++)
	  for(int j=0;j<5;j++)
	    transCount[i][j]=0;

	for(int i=0;i<5;i++)
	  for(int j=0;j<3000;j++)
	    lengthCount[i][j]=0;

}

PriorParam MCEM::start(){
	cerr<<opflist.size()<<" training files in total"<<endl;
	int filecount;
	cerr<<"initial hyperparameters:"<<endl;
	cerr<<prior;

	for(int it=0;it<iteration;it++){
		int hsize=0;
		vector<string>::const_iterator fiter=opflist.begin();
		int itercount=-1;
		filecount=0;
		double marginallogprob=0;
		for(;fiter!=opflist.end();fiter++){
			itercount++;
			string datafname=pathraw+(*fiter);
			cerr<<"processing file:"<<datafname<<endl;
			MaRatioData mrdata(datafname);
			TwoStateSegmentation segmenter(mrdata,args,prior);
			segmenter.forwardRecursion();
			segmenter.computePosteriorKD();
			marginallogprob+=segmenter.marginalLogProb;
		 	cerr<<"posterior probability of K and ending state:"<<endl;
		 	cerr<<segmenter.kmax<<endl;
		 	for (int k=0;k<=segmenter.kmax;k++){
			  cout<<"k="<<k<<" ";
			  for (int state=0;state<prior.statesize;state++)
				  cout<<segmenter.posteriorKD[k][state]<<" ";
			  cout<<endl;
			}
			//REPEAT determines how many samples to draw, defined at the beginning of this file
			for(int repeat=1;repeat<=REPEAT;repeat++){

			  list<int> cutPts,states;
			  double probability=segmenter.backwardSampling(cutPts,states);
  			  hsize+=cutPts.size();
			  //record the information for computing 12 parameter
			  list<int>::const_iterator cIter=cutPts.begin();
			  list<int>::const_iterator sIter=states.begin();
			  int lastcutPts=0;
			  for(;cIter!=cutPts.end();cIter++,sIter++){
				HiddenInfo hinfo;
				hinfo.sum=sum(lastcutPts,*cIter,segmenter.obs);
				hinfo.sumsq=sumsq(lastcutPts,*cIter,segmenter.obs);
				hinfo.state=*sIter;
				hinfo.n=*cIter-lastcutPts+1;
				//if you want to take into account missing data, the above should probably be: 
				//hinfo.n=numobs(lastcutPts, *cIter, segmenter.obs)
				
				recordHiddenInfo(hinfo);
				lastcutPts=*cIter+1;
			  }
			  if (cutPts.size()<=1){
				assert(cutPts.size()==1);
			  }
			  recordSegmentation(cutPts,states);
			}//repeat

		}//for fiter

		//M-step in EM algorithm
		recomputePriorMean();
		recomputePriorTrans();
		recomputePriorD();
		cerr<<"Iteration "<<it<<": marginallogprob:"<<marginallogprob<<endl;
		cerr<<prior;

		cleanRecord();
	}//for it
	return prior;

}

void funfornusig(int n,double x[], double *f){return;}

//log inverse gamma density
double loginvgamma(double sig,double a,double b){
	return -lgamma(a)+a*log(b)-(a+1)*log(sig)-b/sig;
}

//compute value of expected complete likelihood,this is global function
void funfornusig_w_data(int n,double x[], double *f,void *params){
	vector<HiddenInfo> *hinfos=(vector<HiddenInfo> *)params;
	double mu=x[0];
	double k=x[1];
	double nu=x[2];
	double sig=x[3];
	vector<HiddenInfo>::const_iterator hIter=(*hinfos).begin();
	int state=(*hIter).state;
	double res=0.0;
	for(;hIter!=(*hinfos).end();hIter++){
		double sum=(*hIter).sum;
		double sumsq=(*hIter).sumsq;
		int n=(*hIter).n;
		double lambda=(n*k*mu*mu-2*k*mu*sum-sum*sum)/(k+n)+sumsq;

		double newsig=(nu*sig+lambda)/(nu+n);
		//cerr<<"lambda:"<<lambda<<" newsig:"<<newsig<<endl;

			//the 2*pi factor can be omitted
			//temp=0.5*log(prior.k(state))+lg(prior.nu(state),prior.sig(state));
		double save=0.5*log(k)+lg(nu,sig);
		res+=-n/double(2)*log(2*M_PI)+save-0.5*log(k+n)
					-lg(nu+n,newsig);
	}
	*f=-res/REPEAT-loginvgamma(sig,9,9)/REPEAT*((*hinfos).size())-loggammapdf(nu,8,0.5)/REPEAT*((*hinfos).size())
					-loggammapdf(k,20,0.5)/REPEAT*((*hinfos).size());
	//*f=-res-loginvgamma(sig,4,4);
	cerr<<"magnitude:"<<-res/(*hinfos).size()<<" "<<-loginvgamma(sig,0.4,0.4)<<" "<<-loggammapdf(nu,4,4)<<" mu,k,nu,sig "<<mu<<" "<<k<<" "<<nu<<" "<<sig<<" "<<endl;
}


void MCEM::recomputePriorMean(){

	cerr<<"number of segments in total:"<<hiddenInfos[0].size()<<" "<<hiddenInfos[1].size()<<" "<<hiddenInfos[2].size()<<endl;
	cerr<<"more on open state: ";
	for(int i=0;i<hiddenInfos[1].size();i++) cerr<<hiddenInfos[1][i].n<<" ";
	cerr<<endl;
	for(int i=0;i<prior.statesize;i++){
		double xguess[]={prior.mean(i),prior.k(i),prior.nu(i),prior.sig(i)};
		double *a;
		double *b;
		double xlb[]={-1.5,0.01,0.01,0.01};
		double xub[]={2,640,50,15};
		double *x=imsl_d_min_con_gen_lin(funfornusig,4,0,0,a,b,xlb,xub,IMSL_FCN_W_DATA,funfornusig_w_data,(void*)(&hiddenInfos[i]),IMSL_MAX_FCN, 3000,IMSL_XGUESS,xguess,0);
		double prev,postv;
		double pred[5],postd[5];
		
		//print for debugging purposes, can be deleted
		funfornusig_w_data(4,xguess,&prev,(void*)(&hiddenInfos[i]));
		funfornusig_w_data(4,x,&postv,(void*)(&hiddenInfos[i]));
		cerr<<"funvale before:"<<prev<<" funvalue after:"<<postv<<endl;

		//set some hyperparameters to the new values
		prior.setmean(i,x[0]);prior.setk(i,x[1]);prior.setnu(i,x[2]);prior.setsig(i,x[3]);
	}//for i
};

//compute new transition probabilities for M-step
void MCEM::recomputePriorTrans(){
	int sum[5];
	for(int i=0;i<prior.statesize;i++){
		sum[i]=0;
		for(int j=0;j<prior.statesize;j++){
			sum[i]+=transCount[i][j];
		}
	}

	for(int i=0;i<prior.statesize;i++)
	  for(int j=0;j<prior.statesize;j++){
		  double temp=transCount[i][j]/(double)sum[i];
		  if (temp<0.0001) temp=0.0001;
		  prior.settrans(i,j,temp);
	  }
	prior.normalizeTrans();
}

//compute length distributions for M-step
void MCEM::recomputePriorD(){
  for(int i=0;i<prior.statesize;i++){
	double sumcount=0;
	for(int j=SHIFT;j<3000;j++){sumcount+=lengthCount[i][j];}
	double meanlength=0,meansqlength=0;
	for(int j=SHIFT;j<3000;j++){
		meanlength+=(j-SHIFT+1)*(lengthCount[i][j]/sumcount);
		meansqlength+=(j-SHIFT+1)*(j-SHIFT+1)*(lengthCount[i][j]/sumcount);
	}
	double varlength=meansqlength-meanlength*meanlength;
	if (varlength<0.01) { varlength=1; cerr<<"varlength zero:"<<meanlength<<" "<<meansqlength<<endl;}
	//assert(varlength>0);
	prior.setd(i,0,meanlength*meanlength/varlength);
	prior.setd(i,1,varlength/meanlength);
  }
};


void MCEM::recordSegmentation(list<int> cutPts, list<int> states){
	if (cutPts.size()<=1) return;
	list<int>::const_iterator cIter=cutPts.begin();
	list<int>::const_iterator sIter=states.begin();
	int lastcutPts=*cIter;
	int laststate=*sIter;
	cIter++; sIter++;
	int lastlength;
	for(;cIter!=cutPts.end();cIter++,sIter++){
		lengthCount[*sIter][*cIter-lastcutPts]++;
		transCount[laststate][*sIter]++;
		lastlength=*cIter-lastcutPts;
		lastcutPts=*cIter;
		laststate=*sIter;
	}
	lengthCount[laststate][lastlength]--;//do not record the last segment length since it is biased
}


void MCEM::recordHiddenInfo(HiddenInfo hinfo){
	hiddenInfos[hinfo.state].push_back(hinfo);
};

//clean up after each iteration
void MCEM::cleanRecord(){
	for(int i=0;i<5;i++)
	  for(int j=0;j<5;j++)
	    transCount[i][j]=0;

	for(int i=0;i<5;i++)
	  for(int j=0;j<3000;j++)
	    lengthCount[i][j]=0;
    for(int i=0;i<5;i++) hiddenInfos[i].clear();
}


