#ifndef PRIORPARAM_H
#define PRIORPARAM_H

#include <iostream>
#include <fstream>
#include <sstream>
#include <cassert>
//#include <gsl/gsl_linalg.h>
#include "utils.h"
using namespace std;

class PriorParam{
	public:
	PriorParam(){};
	PriorParam(ifstream& is);
	PriorParam(string fname);
	int statesize; //statesize is 1,2,or 3, but the index starts from 0

	//get hyperparameters for the conjugate prior
	double mean(int i) const{assert(i<statesize); return priormean[i];}
	double k(int i) const{assert(i<statesize); return priork[i];}
	double nu(int i) const{assert(i<statesize); return priornu[i];}
	double sig(int i) const{assert(i<statesize); return priorsig[i];}

	//get transition probabilities
	double trans(int i, int j) const
		{assert(i<statesize && j<statesize); return priortrans[i][j];}

	//get parameters for state i and either alpha(0) or beta(1)
	double d(int i,int j) const
	    {assert(i<statesize && j<2); return priord[i][j];}

	//get stationary distribution on states
	double pi(int i) const
		{assert(i<statesize); return priorpi[i];}

	void setmean(int i,double a) {assert(i<statesize); priormean[i]=a;}
	void setk(int i,double a) {assert(i<statesize);  priork[i]=a;}
	void setnu(int i,double a) {assert(i<statesize);  priornu[i]=a;}
	void setsig(int i,double a) {assert(i<statesize);  priorsig[i]=a;}
	void settrans(int i, int j,double a)
		{assert(i<statesize && j<statesize);  priortrans[i][j]=a;}
	void setd(int i,int j,double a)
	    {assert(i<statesize && j<2); priord[i][j]=a;}

	void getpifromtrans();//initial state probability, currently uniform over all states
	void normalizeTrans();//make sure the transition matrix row sum to 1

	friend ostream& operator<<(ostream& os, const PriorParam& prior);

	private:
	double priormean[5];
	double priork[5];
	double priornu[5];
	double priorsig[5];
	double priortrans[5][5];
	double priorpi[5];
	double priord[5][2];
};
#endif
