#include "TwoStateSegmentation.h"
#include "utils.h"
#include <time.h>
#include <stdlib.h>
#include <numeric>

//#define DEBUG

const bool TwoStateSegmentation::uninformative=false;
//currently updateCount has no use, might be used for counting the samples for calculating postmean in the future
TwoStateSegmentation::TwoStateSegmentation(MaRatioData mdata, Args args,PriorParam p):
			mrdata(mdata),updateCount(0),updateVoteCount(0),
			prior(p){
	kmax=40;
	obsLen=mdata.getObsLen();
	obs=vector<double>(obsLen,0.0);//store raw data
	obs=mdata.obs;

        //these values used in recursion to save some computation
	laststart=-1; lastend=-1; lastn=-1;
	lastsum=0.0;lastsumsq=0.0;

	vote=vector<vector<int> >(obsLen,vector<int>(prior.statesize,0));
	centroid=vector<int>(obsLen,0);
	centroidUpperUncertainty=vector<double>(obsLen,0.0);
	centroidLowerUncertainty=vector<double>(obsLen,0.0);

	postMean=vector<double>(obsLen,0.0);
	postVar=vector<double>(obsLen,0.0);
	stdres=vector<double>(obsLen,0.0);

	for(int s=0;s<prior.statesize;s++){
		savecomp[s]=0.5*log(prior.k(s))+lg(prior.nu(s),prior.sig(s));
	}
	/*
	-k number of maximum change points
	-p u/l use either uniform prior on k or Liu&Lawrence prior on k
	*/
	if (args.isset('k')) kmax=atoi(args.value('k').c_str());
	//if (args.isset('p')) computePriorK(args.value('p').c_str());
	probObsDgK=vector<vector<vector<double> > >(obsLen,vector<vector<double> >(prior.statesize,
		vector<double>(kmax+1,0.0)));
	logadjustment=vector<vector<vector<double> > >(obsLen,vector<vector<double> >(kmax+1,
		vector<double>(prior.statesize,0.0)));
	logadjustS=vector<vector<vector<double> > >(obsLen,vector<vector<double> >(kmax+1,
		vector<double>(prior.statesize,0.0)));


	S=vector<vector<vector<double> > >(obsLen,vector<vector<double> >(prior.statesize+1,
		vector<double>(kmax+1,0.0)));
	//zeroS=vector<vector<vector<bool > > >(obsLen,vector<vector<bool> >(prior.statesize+1,
	//	vector<bool>(kmax+1,1)));
	cerr<<"kmax:"<<kmax<<endl;
	//computePriorK();
	//cerr<<"prior of k:"<<priorK[0]<<priorK[1]<<priorK[kmax]<<endl;
	srand48(time(0));
}


//computePriorK() is used only if we use a "pure" change point model with no trans prob or length dist,
//nocurrently implemented
/*
void TwoStateSegmentation::computePriorK(string s){
	priorK=vector<double>(kmax+1);
	if (s=="uniform"){
		for (int i=0;i<=kmax;i++){
			priorK[i]=1/double(kmax+1);
		}
	}
	else
		warning(" right now only uniform prior is allowed");
}*/

//print the log probability of p(obs,delta|k)
/*void TwoStateSegmentation::printProb(int k){
	cerr<<"state 0:"<<endl;
	for(int i=0;i<obsLen;i++) cerr<<probObsDgK[i][0][k]<<" ";
	cerr<<endl;
	cerr<<"state 1:"<<endl;
	for(int i=0;i<obsLen;i++) cerr<<probObsDgK[i][1][k]<<" ";
	cerr<<endl;
}*/

//computeS() is redundent
void TwoStateSegmentation::computeS(){
	cerr<<"computing S....";

	for(int i=0;i<obsLen;i++)
		for (int state=0;state<prior.statesize;state++){
			S[i][state][0]=log(prior.pi(state))+loggammapdf(i+1,prior.d(state,0),prior.d(state,1));
		}
	cerr<<"base case finished"<<endl;
	for(int k=1;k<=kmax;k++){
		cerr<<k<<" ";
		for(int i=k;i<obsLen;i++){
			for(int s=0;s<prior.statesize;s++){
				double X[i][prior.statesize];
				double maxX=-100000000.0;
				for(int j=k-1;j<i;j++){
					for(int pres=0;pres<prior.statesize;pres++){
						X[j][pres]=0.0;
						X[j][pres]=S[j][pres][k-1]-logadjustS[j][k-1][pres]+
									loggammapdf(i-j,prior.d(s,0),prior.d(s,1))+log(prior.trans(pres,s));
						maxX=max(maxX,X[j][pres]);
					}
				}
				//maxX=0;
				logadjustS[i][k][s]=-maxX;
				long double temp=0.0;
				for(int j=k-1;j<i;j++){
					for(int pres=0;pres<prior.statesize;pres++){
						temp+=exp(X[j][pres]-maxX);
					}
				}
				S[i][s][k]=log(temp);
			}
		}
	}
	cerr<<"computeS finished"<<endl;

}



void TwoStateSegmentation::forwardRecursion(){

	computeS();
	double minp=100000.0,maxp=-1000000.0;
	ostringstream ss;
	//initialization for DP, compute prob when there is no change point
	for(int i=0;i<obsLen;i++){
		for(int s=0;s<prior.statesize;s++){
			probObsDgK[i][s][0]=computeOneSegProb(0,i,s,uninformative);
			minp=min(minp,probObsDgK[i][s][0]);
			maxp=max(maxp,probObsDgK[i][s][0]);
		}
	}
	cout<<"maxium and mininum log number after computeOneSeg:"<<minp<<" "<<maxp<<endl;

	minp=100000.0,maxp=-1000000.0;
	for(int i=1;i<obsLen;i++){ //note: i start from 1, since I assume logadjustment at 0 is 0
		for(int s=0;s<prior.statesize;s++)logadjustment[i][0][s]=-probObsDgK[i][s][0];
		for(int s=0;s<prior.statesize;s++){
			probObsDgK[i][s][0]=probObsDgK[i][s][0]+logadjustment[i][0][s];
			minp=min(minp,probObsDgK[i][s][0]);
			maxp=max(maxp,probObsDgK[i][s][0]);
		}
	}
	cout<<"maxium and mininum log number after adjustment:"<<minp<<" "<<maxp<<endl;//should output 0 here

	cerr<<"starting forward recursion,k=";
	for(int k=1;k<=kmax;k++){
		cerr<<k<<" ";
		for(int i=k;i<obsLen;i++)
			for(int state=0; state<prior.statesize;state++){
				long double temp=0.0;
				double X[i-k+1][prior.statesize];
				double maxX=-100000000.0;
				for(int j=i-1;j>=k-1;j--){
					for(int prestate=0;prestate<prior.statesize;prestate++){
						X[j-k+1][prestate]=0.0;
						X[j-k+1][prestate]=probObsDgK[j][prestate][k-1]+computeOneSegProb(j+1,i,state,uninformative)+logadjustment[i][0][state]-logadjustment[j][k-1][prestate]+log(prior.trans(prestate,state))-log(prior.pi(state));
						maxX=max(maxX,X[j-k+1][prestate]);
					}
				}
				for(int j=i-1;j>=k-1;j--){
					for(int prestate=0;prestate<prior.statesize;prestate++){
						temp+=exp(X[j-k+1][prestate]-maxX);
					}
				}
				probObsDgK[i][state][k]=log(temp);
				logadjustment[i][k][state]=logadjustment[i][0][state]-maxX;
				if (isnan(log(temp))||isinf(log(temp))) {
					cerr<<temp<<" "<<maxX<<endl;
					exit(1);
				}

			}//for state
	}
	cerr<<endl;
}

//compute log p(R[start:end],state|no change point),the complication here is due to dealing with
//missing data, and try to save some computation by using previous result
double TwoStateSegmentation::computeOneSegProb(int start,int end,int state,bool diffuse){

	assert(start>=0 && end<obsLen && end>=start);
	double result=0.0;
	  if(end==lastend && abs(start-laststart)<=1){
		//cerr<<"computation saved"<<endl;
		if(start==laststart-1){
			if (lastsum>MISSING && obs[start]>MISSING){
				lastsum+=obs[start];
				lastsumsq+=obs[start]*obs[start];
				lastn=lastn+1;
			}
			if (lastsum<MISSING && obs[start]>MISSING){
				lastsum=obs[start];
				lastsumsq=obs[start]*obs[start];
				lastn=lastn+1;
			}
		}
		if(start==laststart+1){
			if (lastsum>MISSING && obs[laststart]>MISSING){
				lastsum-=obs[laststart];
				lastsumsq-=obs[laststart]*obs[laststart];
			}
			if (lastn==1 && obs[laststart]>MISSING){
				lastsum=double(MISSING-1);
				lastsumsq=double(MISSING-1);
			}
			if (obs[laststart]>MISSING) lastn--;
		}
	  }else if(start==laststart && abs(end-lastend)<=1){
		if(end==lastend-1){
			if(lastsum>MISSING && obs[lastend]>MISSING){
				lastsum-=obs[end+1];
				lastsumsq-=obs[end+1]*obs[end+1];
			}
			if(lastn==1 && obs[lastend]>MISSING){
				lastsum=double(MISSING-1);
				lastsumsq=double(MISSING-1);
			}
			if (obs[lastend]>MISSING) lastn--;
		}
		if(end==lastend+1){
			if (lastsum>MISSING && obs[end]>MISSING){
				lastsum+=obs[end];
				lastsumsq+=obs[end]*obs[end];
				lastn++;
			}
			if (lastsum<MISSING && obs[end]>MISSING){
				lastsum=obs[end];
				lastsumsq=obs[end]*obs[end];
				lastn++;
			}
		}
	  }else{ //if you reach here, that means no way to save computation
		lastsum=sum(start,end,obs);
		lastsumsq=sumsq(start,end,obs);
		lastn=numobs(start,end,obs);
	  }
	  lastend=end;
	  laststart=start;

	  //int n=end-start+1;//if have missing data, this is not truw
	  int n=lastn;//n seems to be just for simplicity of notation
	  //assert(n<=end-start+1)
	  //note: with the following line, this would return p^hat as in my thesis, not p
	  double adjust=log(prior.pi(state))+loggammapdf(end-start+1,prior.d(state,0),prior.d(state,1));

	  if (adjust<-10000000||isnan(adjust)||isinf(adjust)) { cerr<<"adjust"<<prior.pi(state)<<" "<<start<<" "<<end<<" "<<adjust<<endl;exit(1);}

          if (n==0) { result+=adjust; }
	  //look at my thesis for what lambda means
	  double lambda=(n*prior.k(state)*prior.mean(state)*prior.mean(state)
			-2*prior.k(state)*prior.mean(state)*lastsum-lastsum*lastsum)/(prior.k(state)+n)
			+lastsumsq;


	  double res;
	  double newsig=(prior.nu(state)*prior.sig(state)+lambda)/(prior.nu(state)+n);

	  res=-n/double(2)*log(2*M_PI)+savecomp[state]-0.5*log(prior.k(state)+n)
			-lg(prior.nu(state)+n,newsig);

	  if (res<-10000000||isnan(res)) {
		cerr<<"finally"<<adjust<<" "<<newsig<<" "<<obs[105]<<" "<<obs[106]<<endl;exit(1);}
	  result+=res+adjust;

  return result;

}

//compute posterior probability of K and end state
void TwoStateSegmentation::computePosteriorKD(){
	posteriorKD=vector<vector<double> >(kmax+1,vector<double>(prior.statesize,0.0));
	double totalsum=0.0;
	vector<vector<double> >logposteriorKD=vector<vector<double> >(kmax+1,vector<double>(prior.statesize,0.0));
	bool flag=false;//flag used to indicate whether this is first time in calculating maximum
	double maxKD=0.0;
	for(int k=0;k<=kmax;k++){
		for(int state=0;state<prior.statesize;state++){
			logposteriorKD[k][state]=probObsDgK[obsLen-1][state][k]-
					logadjustment[obsLen-1][k][state]+logadjustment[0][k][state];
			if (!flag)
				{maxKD=logposteriorKD[k][state];flag=true;}
			else{
				maxKD=max(maxKD,logposteriorKD[k][state]);
			}
		}
	}
	//convert logprob to prob(unnormalized)
	for(int k=0;k<=kmax;k++)
	for(int state=0;state<prior.statesize;state++){
		posteriorKD[k][state]=exp(logposteriorKD[k][state]-maxKD);
		totalsum+=posteriorKD[k][state];
	}
	marginalLogProb=log(totalsum)+maxKD;//this is the log[P(y)], could be used for model selection
	//normalize probability to sum to 1
	for(int k=0;k<=kmax;k++)
	for(int state=0;state<prior.statesize;state++)
		posteriorKD[k][state]/=totalsum;
}


//draw a single sample, return the log probability
double TwoStateSegmentation::backwardSampling(list<int>& cutPts, list<int>& states){
	double res=0.0;
	double p[(kmax+1)*prior.statesize];

	//compute probability for drawing k and endstate
	for(int k=0;k<=kmax;k++)
		for(int state=0;state<prior.statesize;state++)
			p[state*(kmax+1)+k]=posteriorKD[k][state];

        //draw K and endstate from multinomial dist, convert 1-d index to K and state
	int index=multiSampling(p,(kmax+1)*prior.statesize);
	int sampledK=index%(kmax+1);
	int sampledEndState=index/(kmax+1);

	cutPts.push_front(obsLen-1);
	states.push_front(sampledEndState);
	int currentState=sampledEndState;

	//memorize current probability, to be returned
	res=log(posteriorKD[sampledK][sampledEndState]);

	//recursive backsampling
	for(int k=sampledK;k>0;k--){
		int lastCut=cutPts.front();
		int lastState=states.front();
		double logcutprob[(lastCut-k+1)*prior.statesize];
		double maxp=0.0;
		//if you want to save computation, start with j big, and decrease j,as done below
		bool flag=false;//flag is used to indicate whether this is first in calc of maximum
		for(int j=lastCut-1;j>=k-1;j--){
			for (int prestate=0;prestate<prior.statesize;prestate++){
				logcutprob[j-k+1+prestate*(lastCut-k+1)]=probObsDgK[j][prestate][k-1]+
						computeOneSegProb(j+1,lastCut,lastState,uninformative)-
						logadjustment[j][k-1][prestate]+
						log(prior.trans(prestate,lastState))-log(prior.pi(lastState));

				if (!flag){
					maxp=logcutprob[j-k+1];
					flag=true;
				}
				else{
					maxp=max(maxp,logcutprob[j-k+1+prestate*(lastCut-k+1)]);
				}
			}

		}

		double cutprob[(lastCut-k+1)*prior.statesize];
		double denom=0.0;

		//compute probability,convert logprob to prob
		for(int j=0;j<(lastCut-k+1)*prior.statesize;j++){
			cutprob[j]=0.0;
			//int prestate=j/(lastCut-k+1);
			cutprob[j]=exp(logcutprob[j]-maxp);
			denom+=cutprob[j];
		}
		//normalize to sum to 1
		for(int j=0;j<(lastCut-k+1)*prior.statesize;j++){
			cutprob[j]=cutprob[j]/denom;
		}
		//draw previous changepoint with multinomial
		index=multiSampling(cutprob,(lastCut-k+1)*prior.statesize);
		int changePoint=index%(lastCut-k+1);
		int curstate=index/(lastCut-k+1);
		cutPts.push_front(changePoint+k-1);
		states.push_front(curstate);
		res+=log(cutprob[index]);
	}
	assert(res<=0);
	return res;
}

void TwoStateSegmentation::updatePostmeanvar(const list<int>& cutPts,const list<int>& states){
	//static int updateCount=0;//number of times this update is called
	//cerr<<"updateCount"<<updateCount<<endl;
	//cerr<<"postMean:"<<postMean[10]<<postMean[25]<<postMean[44]<<endl;
	assert(cutPts.size()==states.size());
	list<int>::const_iterator cIter=cutPts.begin();
	list<int>::const_iterator sIter=states.begin();
	int currentc=*cIter;
	int currents=*sIter;
	int precut=-1;
	//vector<double> pmtemp=vector<double>(obsLen,0.0);
	for(;cIter!=cutPts.end();cIter++,sIter++){
		currentc=*cIter;
		currents=*sIter;
		assert(precut<currentc);

		double sumobs=sum(precut+1,currentc,obs);
		int sumn=numobs(precut+1,currentc,obs);
		double s2=variance(precut+1,currentc,obs);
		if  (s2<0) { assert(s2<MISSING); s2=0;}

		double newmean=(prior.k(currents)*prior.mean(currents)+sumobs)/(prior.k(currents)+sumn);
		//newvar computes nu_n sig_n^2 on page 72 of Gelman
		double newvar=prior.nu(currents)*prior.sig(currents)+(sumn-1)*s2+
				prior.k(currents)*sumn*pow((sumobs/sumn-prior.mean(currents)),2)/(prior.k(currents)+sumn);
		//if (isinf(newvar)) {cerr<<"shouldn't be"<<endl; exit(1);}
		newvar=newvar/(prior.nu(currents)+sumn-2);//now newvar is mean of the posterior variance
		//if (isinf(newvar)) {cerr<<sumn<<endl; exit(1);}
		for(int i=precut+1;i<=currentc;i++){
			if (obs[i]<MISSING) continue;
			postMean[i]=(postMean[i]*updateCount+newmean)/(updateCount+1);
			postVar[i]=(postVar[i]*updateCount+newvar)/(updateCount+1);
		}
		precut=currentc;
	}
	updateCount++;
}

void TwoStateSegmentation::cleanPostmeanvar(){
	for(int i=0;i<obsLen;i++){
		postMean[i]=0;postVar[i]=0;
	}
	updateCount=0;
}


//store vote from a single sample in vote[][]
void TwoStateSegmentation::updateVote(const list<int>& cutPts,const list<int>& states){
	updateVoteCount++;
	assert(cutPts.size()==states.size());
	vector<int> fun=convertSegToFun(cutPts,states);
	for(int i=0;i<fun.size();i++)	vote[i][fun[i]]++;
}

//for each position, find the state assignment by majority vote
void TwoStateSegmentation::computeCentroid(){
	for(int i=0;i<obsLen;i++){
		int currentvote=-1;
		int currents=-1;
		for(int s=0;s<prior.statesize;s++){
			if (vote[i][s]>currentvote){
				currentvote=vote[i][s];
				currents=s;
			}
		}
		centroid[i]=currents;
	}
}

//computeUncertainty should be called after computeCentroid(),this is not currently used,
//certainty is actually computed from centroid, refer to ::printCertainty
void TwoStateSegmentation::computeUncertainty(){
	cerr<<"computing uncertainty"<<endl;
	for(int i=0;i<obsLen;i++){
		for(int s=0;s<prior.statesize;s++){
			if (s>centroid[i])
				centroidUpperUncertainty[i]+=vote[i][s]*(s-centroid[i]);
			else if (s<centroid[i])
				centroidLowerUncertainty[i]+=vote[i][s]*(centroid[i]-s);
		}
		centroidUpperUncertainty[i]/=prior.statesize*updateVoteCount/1.5;
		centroidLowerUncertainty[i]/=prior.statesize*updateVoteCount/1.5;
	}
}

void TwoStateSegmentation::computeStdres(){
   for(int i=0;i<obsLen;i++){
	if (obs[i]<MISSING) continue;
	stdres[i]=(obs[i]-postMean[i])/sqrt(postVar[i]);
   }
}  

//compute sum of 0-1 distance between two state assignment, not currently used, provided
//for convenience or future use
int TwoStateSegmentation::distance(const vector<int>& fun1, const vector<int>& fun2){
	assert(fun1.size()==fun2.size());
	int dis=0;
	for(int i=0;i<fun1.size();i++){
		if (fun1[i]!=fun2[i]) dis++;
	}
	return dis;
}

//convert segmentation information to state assignments at each point, used by ::updateVote()
vector<int> TwoStateSegmentation::convertSegToFun(const list<int>& cutPts,const list<int>& states){
	assert(cutPts.size()==states.size());
	int length=cutPts.back()+1;
	vector<int> fun(length,0);
	list<int>::const_iterator cIter=cutPts.begin();
	list<int>::const_iterator sIter=states.begin();
	for(int i=0;i<length;i++){
		fun[i]=*sIter;
		if (i>=*cIter){
			cIter++;
			sIter++;
		}
	}
	return fun;
}



void TwoStateSegmentation::printObs(ostream& out){
	for(int i=0;i<obsLen;i++){
	  if (obs[i]<MISSING) continue;
	  Site satdata=mrdata.data[i];
	  out<<satdata.regionName<<"\t"<<satdata.start<<"\t"<<satdata.end-12<<"\t"<<obs[i]<<endl;
	}
}

void TwoStateSegmentation::printCentroid(ostream& out){
	for(int i=0;i<obsLen;i++){
		if (obs[i]<MISSING) continue;
		Site satdata=mrdata.data[i];
		out<<satdata.regionName<<"\t"<<satdata.start<<"\t"<<satdata.end-12<<"\t";
		out<<"\t"<<centroid[i]<<endl;
	}
}

void TwoStateSegmentation::printPostmean(ostream& out){
	for(int i=0;i<obsLen;i++){
		if (obs[i]<MISSING) continue;
		Site satdata=mrdata.data[i];
		out<<satdata.regionName<<"\t"<<satdata.start<<"\t"<<satdata.end-12<<"\t";
		out<<"\t"<<postMean[i]<<endl;
	}
}

void TwoStateSegmentation::printPostvar(ostream& out){
	for(int i=0;i<obsLen;i++){
		if (obs[i]<MISSING) continue;
		Site satdata=mrdata.data[i];
		out<<satdata.regionName<<"\t"<<satdata.start<<"\t"<<satdata.end-12<<"\t";
		out<<"\t"<<postVar[i]<<endl;
	}
}

void TwoStateSegmentation::printStdres(ostream& out){
	for(int i=0;i<obsLen;i++){
		if (obs[i]<MISSING) continue;
		Site satdata=mrdata.data[i];
		out<<satdata.regionName<<"\t"<<satdata.start<<"\t"<<satdata.end-12<<"\t";
		out<<"\t"<<stdres[i]<<endl;
	}
}


void TwoStateSegmentation::printCertainty(ostream& out){
	for(int i=0;i<obsLen;i++){
		if (obs[i]<MISSING) continue;
		Site satdata=mrdata.data[i];
		out<<satdata.regionName<<"\t"<<satdata.start<<"\t"<<satdata.end-12<<"\t";
		out<<"\t"<<vote[i][centroid[i]]/double(updateVoteCount)<<endl;
	}
}







