#ifndef TWOSTATESEGMENTATION_H
#define TWOSTATESEGMENTATION_H

#include <iostream>
#include <fstream>
#include <math.h>
#include <assert.h>
#include <vector>
#include <string>
#include "Args.h"
#include "MaRatioData.h"
#include "utils.h"
#include "PriorParam.h"
#include "HiddenInfo.h"


using namespace std;
class TwoStateSegmentation{
	public:
	TwoStateSegmentation(MaRatioData mdata, Args args, PriorParam p);
	MaRatioData mrdata;
	vector<double> obs;
	vector<int> centroid;
	vector<double> centroidUpperUncertainty;
	vector<double> centroidLowerUncertainty;
	vector<vector<int> > vote;
	vector<double> postMean;
	vector<double> postVar;
	vector<double> stdres;
	int kmax;
	//int maxSegLen;
	//int minSegLen;
	int obsLen;
	double marginalLogProb;
	vector<double> priorK;
	vector<vector<double> > posteriorKD;
	vector<vector<vector<double> > > probObsDgK;
	vector<vector<vector<double> > > S;
	void forwardRecursion();
	double backwardSampling(list<int>& cutPts,list<int>& states);
	void computePosteriorKD();
	void computeS();
	void updateVote(const list<int>& cutPts,const list<int>& states);
	void updatePostmeanvar(const list<int>& cutPts, const list<int>& states);
	void cleanPostmeanvar();
	void computeCentroid();
	void computeUncertainty();
	void computeStdres();
	void printCentroid(ostream& out);
	void printCertainty(ostream& out);
	void printObs(ostream& out);
	void printPostmean(ostream& out);
	void printPostvar(ostream& out);
	void printStdres(ostream& out);

	static vector<int> convertSegToFun(const list<int>& cutPts,const list<int>& states);
	static int distance(const vector<int>& fun1, const vector<int>& fun2);
	private:
	PriorParam prior;
	static const bool uninformative;//no use now
	int updateCount;//in the future, might be used by updatePostMean to keep track of the number of samples
	int updateVoteCount;


	//used by computeOneSegProb to save computation
	int laststart,lastend,lastn;
	double lastsum,lastsumsq;
	double savecomp[5];

	vector<vector<vector<double> > >logadjustment;
	vector<vector<vector<double> > >logadjustS;

	double computeOneSegProb(int start,int end,int state,bool diffuse=false);

	void printProb(int k);
};

#endif
