#include "utils.h"


void warning(const string& message){
	cerr<<"!!!Warning:"<<message<<endl;
}

//right now this is only used when compute posterior of k
double lognchoosek(int n, int k){
	assert(n>=0 && k>=0 && k<=n);
	return lgamma(n+1)-lgamma(k+1)-lgamma(n-k+1);
}

string int2string(int i){
	char temp[16];
	sprintf(temp,"%i",i);
	return string(temp);
}

//look at my thesis for function g(.,.), this computes its log
double lg(double nu,double sig){
	return nu/2*log(nu/2)-lgamma(nu/2)+nu*log(sig);
}

//return the range of j such that j>=0,j<i,and i-j \in [lower,upper]
//this will be used when I incorporate simpler model in the future
pair<int,int> range(int i,int lower,int upper){
	assert(i>0);
	if (upper<0) upper=i;
	if (lower<=0) lower=1;
	int l=(upper<=i? i-upper:0);
	int u=i-lower;
	if (u<l) return make_pair(-1,-1);
	else return make_pair(l,u);
}

//this not used now
bool inRange(int j,int i,int lower, int upper){
	assert(j>=0);
	pair<int,int> p=range(i,lower,upper);
	return (j>=p.first && j<=p.second);
}

//might be used during debugging
void monitor(double d,const string& message){
	ostringstream ss;
	if (abs((int)d)>pow(10.0,100)){
		//assert(0);
		ss<<"numbers got too big"<<message<<":"<<d;
		warning(ss.str());
		exit(0);
	}
	else if(abs((int)d)<pow(10.0,-100)){
		//assert(0);
		ss<<"numbers got too small"<<message<<":"<<d;
		warning(ss.str());
		exit(0);
	}
}

int multiSampling(double* prob, int size){
	//srand should be called by the class that use this utility function
	double r=drand48();
	double temp=0.0;
	int index=0;
	double totalsum=0.0;
	for(int i=0;i<size;i++) totalsum+=prob[i];
	for(;index<size;index++){
		temp=temp+prob[index]/totalsum;
		if (temp>r) break;
	}
	if(index>=size){
		cerr<<"error inside multiSampling"<<endl;
		for(int i=0;i<size;i++)	cerr<<prob[i]<<" ";
		cerr<<"totalsum(should be 1):"<<totalsum<<endl;
		assert(index<size);
	}
	return index;
}

//utility to get rid of comment at the end of line
string ridComment(const string& s){
	//assumes that s contains trailing comment
	int i=s.find("#");
	if (i==string::npos){
		return s;
	}
	else{ return s.substr(0,i);}
}

//compute log gamma density, SHIFT shift distribution to the right, thus provide lower
//bound for segment length, SHIFT is defined in utils.h
double loggammapdf(double x,double alpha,double beta){
	assert(x>0 && beta>0);
	//if (x<100) return log(0.01);
	//return -10000;
	//return (x-1)*log(0.9)+log(0.1);
	if (x<SHIFT-0.001) return -1000;
	x=x-SHIFT+1;
	double res=(alpha-1)*log(x)-x/beta-alpha*log(beta)-lgamma(alpha);
	return res;
}

